@@ -25,6 +25,8 @@
 #include <private/qv4generatorobject_p.h>
 #include <private/qv4resolvedtypereference_p.h>
 #include <private/qqmlpropertybinding_p.h>
+#include <private/qqmlanybinding_p.h>
+#include <QtQml/private/qqmlvme_p.h>
 
 #include <QScopedValueRollback>
 
@@ -36,6 +38,19 @@ Q_LOGGING_CATEGORY(lcQmlDefaultMethod, "qt.qml.defaultmethod")
 
 QT_USE_NAMESPACE
 
+Q_TRACE_PREFIX(qtqml,
+"namespace QV4 {" \
+"struct ExecutionEngine;" \
+"namespace CompiledData {" \
+"struct CompilationUnit;" \
+"struct Object;" \
+"}}" \
+"class QQmlEngine;"
+)
+
+Q_TRACE_POINT(qtqml, QQmlObjectCreator_createInstance_entry, const QV4::CompiledData::CompilationUnit *compilationUnit, const QV4::CompiledData::Object *object, const QUrl &url)
+Q_TRACE_POINT(qtqml, QQmlObjectCreator_createInstance_exit, const QString &typeName)
+
 QQmlObjectCreator::QQmlObjectCreator(
         QQmlRefPointer<QQmlContextData> parentContext,
         const QQmlRefPointer<QV4::ExecutableCompilationUnit> &compilationUnit,
@@ -275,6 +290,16 @@ void QQmlObjectCreator::populateDeferredBinding(const QQmlProperty &qmlProperty,
     }
 }
 
+void QQmlObjectCreator::populateDeferredInstance(
+        QObject *outerObject, int deferredIndex, int index, QObject *instance,
+        QObject *bindingTarget, const QQmlPropertyData *valueTypeProperty,
+        const QV4::CompiledData::Binding *binding)
+{
+    doPopulateDeferred(outerObject, deferredIndex, [&]() {
+        populateInstance(index, instance, bindingTarget, valueTypeProperty, binding);
+    });
+}
+
 void QQmlObjectCreator::finalizePopulateDeferred()
 {
     phase = ObjectsCreated;
@@ -424,9 +449,9 @@ void QQmlObjectCreator::setPropertyValue(const QQmlPropertyData *property, const
     }
     break;
     case QMetaType::QColor: {
-        QVariant data;
-        if (QQml_valueTypeProvider()->createValueType(
-                    propertyType, compilationUnit->bindingValueAsString(binding), data)) {
+        QVariant data = QQmlValueTypeProvider::createValueType(
+                    compilationUnit->bindingValueAsString(binding), propertyType);
+        if (data.isValid()) {
             property->writeProperty(_qobject, data.data(), propertyWriteFlags);
         }
     }
@@ -507,11 +532,9 @@ void QQmlObjectCreator::setPropertyValue(const QQmlPropertyData *property, const
     case QMetaType::QVector3D:
     case QMetaType::QVector4D:
     case QMetaType::QQuaternion: {
-        QVariant result;
-        bool ok = QQml_valueTypeProvider()->createValueType(
-                    propertyType, compilationUnit->bindingValueAsString(binding), result);
-        assertOrNull(ok);
-        Q_UNUSED(ok);
+        QVariant result = QQmlValueTypeProvider::createValueType(
+                    compilationUnit->bindingValueAsString(binding), propertyType);
+        assertOrNull(result.isValid());
         property->writeProperty(_qobject, result.data(), propertyWriteFlags);
         break;
     }
@@ -575,6 +598,35 @@ void QQmlObjectCreator::setPropertyValue(const QQmlPropertyData *property, const
             }
             property->writeProperty(_qobject, &value, propertyWriteFlags);
             break;
+        } else {
+            QVariant source;
+            switch (binding->type()) {
+            case QV4::CompiledData::Binding::Type_Boolean:
+                source = binding->valueAsBoolean();
+                break;
+            case QV4::CompiledData::Binding::Type_Number: {
+                const double n = compilationUnit->bindingValueAsNumber(binding);
+                if (double(int(n)) == n)
+                    source = int(n);
+                else
+                    source = n;
+                break;
+            }
+            case QV4::CompiledData::Binding::Type_Null:
+                source = QVariant::fromValue<std::nullptr_t>(nullptr);
+                break;
+            case QV4::CompiledData::Binding::Type_Invalid:
+                break;
+            default:
+                source = compilationUnit->bindingValueAsString(binding);
+                break;
+            }
+
+            QVariant target = QQmlValueTypeProvider::createValueType(source, propertyType);
+            if (target.isValid()) {
+                property->writeProperty(_qobject, target.data(), propertyWriteFlags);
+                break;
+            }
         }
 
         // string converters are not exposed, so ending up here indicates an error
@@ -663,8 +715,9 @@ void QQmlObjectCreator::setupBindings(BindingSetupFlags mode)
                 QQmlData *data = QQmlData::get(targetObject);
                 Q_ASSERT(data && data->propertyCache);
                 targetProperty = data->propertyCache->property(targetIndex.coreIndex());
+                sharedState->requiredProperties.remove({targetObject, targetProperty});
             }
-            sharedState->requiredProperties.remove(targetProperty);
+            sharedState->requiredProperties.remove({_bindingTarget, property});
         }
 
 
@@ -817,8 +870,18 @@ bool QQmlObjectCreator::setPropertyBinding(const QQmlPropertyData *bindingProper
                 void *argv[1] = { &groupObject };
                 QMetaObject::metacall(_qobject, QMetaObject::ReadProperty, bindingProperty->coreIndex(), argv);
                 if (!groupObject) {
-                    recordError(binding->location, tr("Cannot set properties on %1 as it is null").arg(stringAt(binding->propertyNameIndex)));
-                    return false;
+                    QQmlPropertyIndex index(bindingProperty->coreIndex());
+                    auto anyBinding = QQmlAnyBinding::ofProperty(_qobject, index);
+                    if (anyBinding) {
+                        // if there is a binding, try to force-evaluate it now
+                        // this might instantiate a necessary part of a grouped property
+                        anyBinding.refresh();
+                        QMetaObject::metacall(_qobject, QMetaObject::ReadProperty, bindingProperty->coreIndex(), argv);
+                    }
+                    if (!groupObject) {
+                        recordError(binding->location, tr("Cannot set properties on %1 as it is null").arg(stringAt(binding->propertyNameIndex)));
+                        return false;
+                    }
                 }
 
                 bindingTarget = groupObject;
@@ -846,14 +909,13 @@ bool QQmlObjectCreator::setPropertyBinding(const QQmlPropertyData *bindingProper
             && !(bindingFlags & QV4::CompiledData::Binding::IsPropertyObserver)
             && !_valueTypeProperty;
 
-    if (_ddata->hasBindingBit(bindingProperty->coreIndex()) && allowedToRemoveBinding) {
-        QQmlPropertyPrivate::removeBinding(_bindingTarget, QQmlPropertyIndex(bindingProperty->coreIndex()));
-    } else if (bindingProperty->isBindable() && allowedToRemoveBinding) {
-        QList<DeferredQPropertyBinding> &pendingBindings = sharedState.data()->allQPropertyBindings;
-        auto it = std::remove_if(pendingBindings.begin(), pendingBindings.end(), [&](const DeferredQPropertyBinding &deferred) {
-            return deferred.properyIndex == bindingProperty->coreIndex() && deferred.target == _bindingTarget;
-        });
-        pendingBindings.erase(it, pendingBindings.end());
+    if (allowedToRemoveBinding) {
+        if (bindingProperty->isBindable()) {
+            removePendingBinding(_bindingTarget, bindingProperty->coreIndex());
+        } else {
+            QQmlPropertyPrivate::removeBinding(
+                    _bindingTarget, QQmlPropertyIndex(bindingProperty->coreIndex()));
+        }
     }
 
     if (bindingType == QV4::CompiledData::Binding::Type_Script || binding->isTranslationBinding()) {
@@ -899,6 +961,9 @@ bool QQmlObjectCreator::setPropertyBinding(const QQmlPropertyData *bindingProper
                 qmlBinding = QQmlPropertyBinding::create(bindingProperty, runtimeFunction, _scopeObject, context, currentQmlContext(), _bindingTarget, index);
             }
             sharedState.data()->allQPropertyBindings.push_back(DeferredQPropertyBinding {_bindingTarget, bindingProperty->coreIndex(), qmlBinding });
+
+            QQmlData *data = QQmlData::get(_bindingTarget, true);
+            data->setBindingBit(_bindingTarget, bindingProperty->coreIndex());
         } else {
             // When writing bindings to grouped properties implemented as value types,
             // such as point.x: { someExpression; }, then the binding is installed on
@@ -1182,21 +1247,12 @@ QObject *QQmlObjectCreator::createInstance(int index, QObject *parent, bool isCo
         if (type.isValid() && !type.isInlineComponentType()) {
             typeName = type.qmlTypeName();
 
-            void *ddataMemory = nullptr;
-            instance = type.create(&ddataMemory, sizeof(QQmlData));
+            instance = type.createWithQQmlData();
             if (!instance) {
                 recordError(obj->location, tr("Unable to create object of type %1").arg(stringAt(obj->inheritedTypeNameIndex)));
                 return nullptr;
             }
 
-            {
-                QQmlData *ddata = new (ddataMemory) QQmlData;
-                ddata->ownMemory = false;
-                QObjectPrivate* p = QObjectPrivate::get(instance);
-                Q_ASSERT(!p->isDeletingChildren);
-                p->declarativeData = ddata;
-            }
-
             const int finalizerCast = type.finalizerCast();
             if (finalizerCast != -1) {
                 auto hook = reinterpret_cast<QQmlFinalizerHook *>(reinterpret_cast<char *>(instance) + finalizerCast);
@@ -1235,11 +1291,26 @@ QObject *QQmlObjectCreator::createInstance(int index, QObject *parent, bool isCo
                     return nullptr;
                 }
             } else {
-                int subObjectId = type.inlineComponentId();
-                QScopedValueRollback<int> rollback {compilationUnit->icRoot, subObjectId};
+                QString subObjectName;
+                if (compilationUnit->icRootName) {
+                    subObjectName = type.elementName();
+                    std::swap(*compilationUnit->icRootName, subObjectName);
+                } else {
+                    compilationUnit->icRootName = std::make_unique<QString>(type.elementName());
+                }
+
+                const auto guard = qScopeGuard([&] {
+                    if (subObjectName.isEmpty())
+                        compilationUnit->icRootName.reset();
+                    else
+                        std::swap(*compilationUnit->icRootName, subObjectName);
+                });
+
                 QQmlObjectCreator subCreator(context, compilationUnit, sharedState.data(),
                                              isContextObject);
-                instance = subCreator.create(subObjectId, nullptr, nullptr, CreationFlags::InlineComponent);
+                instance = subCreator.create(
+                    compilationUnit->inlineComponentId(*compilationUnit->icRootName),
+                    nullptr, nullptr, CreationFlags::InlineComponent);
                 if (!instance) {
                     errors += subCreator.errors;
                     return nullptr;
@@ -1404,17 +1475,27 @@ bool QQmlObjectCreator::finalize(QQmlInstantiationInterrupt &interrupt)
 
     while (!sharedState->allQPropertyBindings.isEmpty()) {
         auto& [target, index, qmlBinding] = sharedState->allQPropertyBindings.first();
+
+        QQmlData *data = QQmlData::get(target);
+        if (!data || !data->hasBindingBit(index)) {
+            // The target property has been overwritten since we stashed the binding.
+            sharedState->allQPropertyBindings.pop_front();
+            continue;
+        }
+
         QUntypedBindable bindable;
         void *argv[] = { &bindable };
         // allow interception
         target->metaObject()->metacall(target, QMetaObject::BindableProperty, index, argv);
         const bool success = bindable.setBinding(qmlBinding);
 
+        const auto bindingPrivateRefCount = QPropertyBindingPrivate::get(qmlBinding)->ref;
+
         // Only pop_front after setting the binding as the bindings are refcounted.
         sharedState->allQPropertyBindings.pop_front();
 
         // If the binding was actually not set, it's deleted now.
-        if (success) {
+        if (success && bindingPrivateRefCount > 1) {
             if (auto priv = QPropertyBindingPrivate::get(qmlBinding); priv->hasCustomVTable()) {
                 auto qmlBindingPriv = static_cast<QQmlPropertyBinding *>(priv);
                 auto jsExpression = qmlBindingPriv->jsExpression();
@@ -1549,7 +1630,7 @@ bool QQmlObjectCreator::populateInstance(int index, QObject *instance, QObject *
             postHocRequired.erase(postHocIt);
         if (isContextObject)
             sharedState->hadTopLevelRequiredProperties = true;
-        sharedState->requiredProperties.insert(propertyData,
+        sharedState->requiredProperties.insert({_qobject, propertyData},
                                                RequiredPropertyInfo {compilationUnit->stringAt(property->nameIndex), compilationUnit->finalUrl(), property->location, {}});
 
     }
@@ -1609,7 +1690,7 @@ bool QQmlObjectCreator::populateInstance(int index, QObject *instance, QObject *
         if (isContextObject)
             sharedState->hadTopLevelRequiredProperties = true;
         sharedState->requiredProperties.insert(
-                propertyData,
+                {_qobject, propertyData},
                 RequiredPropertyInfo {
                         name, compilationUnit->finalUrl(), _compiledObject->location, {} });
     }
@@ -1640,7 +1721,7 @@ bool QQmlObjectCreator::populateInstance(int index, QObject *instance, QObject *
             if (isContextObject)
                 sharedState->hadTopLevelRequiredProperties = true;
             sharedState->requiredProperties.insert(
-                    propertyData,
+                    {_qobject, propertyData},
                     RequiredPropertyInfo {
                             name, compilationUnit->finalUrl(), _compiledObject->location, {} });
         }
@@ -1673,7 +1754,7 @@ bool QQmlObjectCreator::populateInstance(int index, QObject *instance, QObject *
         const QQmlPropertyData *const targetProperty = targetDData->propertyCache->property(coreIndex);
         if (!targetProperty)
             continue;
-        auto it = sharedState->requiredProperties.find(targetProperty);
+        auto it = sharedState->requiredProperties.find({target, targetProperty});
         if (it != sharedState->requiredProperties.end())
             it->aliasesToRequired.push_back(
                     AliasToRequiredInfo {
